@@ -32,10 +32,11 @@
 #include "acceptdata.h"
 #include "zhttpmanager.h"
 #include "zhttprequest.h"
-#include "domainmap.h"
+#include "zroutes.h"
 #include "xffrule.h"
 #include "requestsession.h"
 #include "proxyutil.h"
+#include "acceptrequest.h"
 
 #define MAX_ACCEPT_REQUEST_BODY 100000
 #define MAX_ACCEPT_RESPONSE_BODY 100000
@@ -83,18 +84,19 @@ public:
 
 	ProxySession *q;
 	State state;
+	ZRoutes *zroutes;
 	ZhttpManager *zhttpManager;
-	DomainMap *domainMap;
 	ZhttpRequest *inRequest;
+	ZrpcManager *acceptManager;
 	bool isHttps;
-	QByteArray routeId;
-	QByteArray channelPrefix;
+	DomainMap::Entry route;
 	QList<DomainMap::Target> targets;
 	ZhttpRequest *zhttpRequest;
 	bool addAllowed;
 	bool haveInspectData;
 	InspectData idata;
-	QSet<QByteArray> acceptTypes;
+	QSet<QByteArray> acceptHeaderPrefixes;
+	QSet<QByteArray> acceptContentTypes;
 	QSet<SessionItem*> sessionItems;
 	HttpRequestData requestData;
 	HttpResponseData responseData;
@@ -113,14 +115,16 @@ public:
 	XffRule xffRule;
 	XffRule xffTrustedRule;
 	QList<QByteArray> origHeadersNeedMark;
+	AcceptRequest *acceptRequest;
 
-	Private(ProxySession *_q, ZhttpManager *_zhttpManager, DomainMap *_domainMap) :
+	Private(ProxySession *_q, ZRoutes *_zroutes, ZrpcManager *_acceptManager) :
 		QObject(_q),
 		q(_q),
 		state(Stopped),
-		zhttpManager(_zhttpManager),
-		domainMap(_domainMap),
+		zroutes(_zroutes),
+		zhttpManager(0),
 		inRequest(0),
+		acceptManager(_acceptManager),
 		isHttps(false),
 		zhttpRequest(0),
 		addAllowed(true),
@@ -128,9 +132,11 @@ public:
 		requestBytesToWrite(0),
 		total(0),
 		passToUpstream(false),
-		useXForwardedProtocol(false)
+		useXForwardedProtocol(false),
+		acceptRequest(0)
 	{
-		acceptTypes += "application/grip-instruct";
+		acceptHeaderPrefixes += "Grip-";
+		acceptContentTypes += "application/grip-instruct";
 	}
 
 	~Private()
@@ -141,15 +147,27 @@ public:
 	void cleanup()
 	{
 		foreach(SessionItem *si, sessionItems)
+		{
+			// emitting a signal here is gross, but this way the engine cleans up the request sessions
+			emit q->requestSessionDestroyed(si->rs, false);
 			delete si->rs;
+			delete si;
+		}
 
 		sessionItems.clear();
 		sessionItemsBySession.clear();
+
+		if(zhttpManager)
+		{
+			zroutes->removeRef(zhttpManager);
+			zhttpManager = 0;
+		}
 	}
 
 	void add(RequestSession *rs)
 	{
 		assert(addAllowed);
+		assert(!route.isNull());
 
 		SessionItem *si = new SessionItem;
 		si->rs = rs;
@@ -170,30 +188,22 @@ public:
 			requestBody += requestData.body;
 			requestData.body.clear();
 
-			DomainMap::Entry entry = domainMap->entry(DomainMap::Http, isHttps, host, requestData.uri.encodedPath());
-			if(entry.isNull())
-			{
-				log_warning("proxysession: %p %s has 0 routes", q, qPrintable(host));
-				rejectAll(502, "Bad Gateway", QString("No route for host: %1").arg(host));
-				return;
-			}
+			if(!route.asHost.isEmpty())
+				requestData.uri.setHost(route.asHost);
 
-			if(!entry.asHost.isEmpty())
-				requestData.uri.setHost(entry.asHost);
-
-			if(entry.pathRemove > 0)
+			if(route.pathRemove > 0)
 			{
 				QByteArray path = requestData.uri.encodedPath();
-				path = path.mid(entry.pathRemove);
+				path = path.mid(route.pathRemove);
 				requestData.uri.setEncodedPath(path);
 			}
 
 			QByteArray sigIss;
 			QByteArray sigKey;
-			if(!entry.sigIss.isEmpty() && !entry.sigKey.isEmpty())
+			if(!route.sigIss.isEmpty() && !route.sigKey.isEmpty())
 			{
-				sigIss = entry.sigIss;
-				sigKey = entry.sigKey;
+				sigIss = route.sigIss;
+				sigKey = route.sigKey;
 			}
 			else
 			{
@@ -201,13 +211,9 @@ public:
 				sigKey = defaultSigKey;
 			}
 
-			routeId = entry.id;
-			channelPrefix = entry.prefix;
-			targets = entry.targets;
+			targets = route.targets;
 
-			log_debug("proxysession: %p %s has %d routes", q, qPrintable(host), targets.count());
-
-			bool trustedClient = ProxyUtil::manipulateRequestHeaders("wsproxysession", q, &requestData, defaultUpstreamKey, entry, sigIss, sigKey, useXForwardedProtocol, xffTrustedRule, xffRule, origHeadersNeedMark, rs->peerAddress());
+			bool trustedClient = ProxyUtil::manipulateRequestHeaders("wsproxysession", q, &requestData, defaultUpstreamKey, route, sigIss, sigKey, useXForwardedProtocol, xffTrustedRule, xffRule, origHeadersNeedMark, rs->peerAddress(), idata);
 
 			if(trustedClient)
 				passToUpstream = true;
@@ -283,7 +289,21 @@ public:
 		if(!target.host.isEmpty())
 			uri.setHost(target.host);
 
-		log_debug("proxysession: %p forwarding to %s:%d", q, qPrintable(target.connectHost), target.connectPort);
+		if(zhttpManager)
+			zroutes->removeRef(zhttpManager);
+
+		if(target.type == DomainMap::Target::Custom)
+		{
+			zhttpManager = zroutes->managerForRoute(target.zhttpRoute);
+			log_debug("proxysession: %p forwarding to %s", q, qPrintable(target.zhttpRoute.baseSpec));
+		}
+		else // Default
+		{
+			zhttpManager = zroutes->defaultManager();
+			log_debug("proxysession: %p forwarding to %s:%d", q, qPrintable(target.connectHost), target.connectPort);
+		}
+
+		zroutes->addRef(zhttpManager);
 
 		zhttpRequest = zhttpManager->createRequest();
 		zhttpRequest->setParent(this);
@@ -297,8 +317,11 @@ public:
 		if(target.insecure)
 			zhttpRequest->setIgnoreTlsErrors(true);
 
-		zhttpRequest->setConnectHost(target.connectHost);
-		zhttpRequest->setConnectPort(target.connectPort);
+		if(target.type == DomainMap::Target::Default)
+		{
+			zhttpRequest->setConnectHost(target.connectHost);
+			zhttpRequest->setConnectPort(target.connectPort);
+		}
 
 		zhttpRequest->start(requestData.method, uri, requestData.headers);
 
@@ -365,6 +388,21 @@ public:
 		}
 	}
 
+	void respondAll(int code, const QByteArray &reason, const HttpHeaders &headers, const QByteArray &body)
+	{
+		foreach(SessionItem *si, sessionItems)
+		{
+			if(si->state != SessionItem::Errored)
+			{
+				assert(si->state == SessionItem::WaitingForResponse);
+
+				si->state = SessionItem::Responded;
+				si->bytesToWrite = -1;
+				si->rs->respond(code, reason, headers, body);
+			}
+		}
+	}
+
 	void destroyAll()
 	{
 		// this method is only to be called when we are in Responding state
@@ -479,11 +517,17 @@ public:
 
 			if(state == Accepting)
 			{
-				foreach(SessionItem *si, sessionItems)
+				if(acceptManager)
 				{
-					si->state = SessionItem::Pausing;
-					si->rs->pause();
+					log_debug("we have an acceptmanager");
+					foreach(SessionItem *si, sessionItems)
+					{
+						si->state = SessionItem::Pausing;
+						si->rs->pause();
+					}
 				}
+				else
+					cannotAcceptAll();
 			}
 			else // Responding
 			{
@@ -531,12 +575,38 @@ public slots:
 			total += responseBody.size();
 			log_debug("proxysession: %p recv total: %d", q, total);
 
-			QByteArray contentType = responseData.headers.get("Content-Type");
-			int at = contentType.indexOf(';');
-			if(at != -1)
-				contentType = contentType.mid(0, at);
+			bool doAccept = false;
+			if(!passToUpstream)
+			{
+				QByteArray contentType = responseData.headers.get("Content-Type");
+				int at = contentType.indexOf(';');
+				if(at != -1)
+					contentType = contentType.mid(0, at);
 
-			if(!passToUpstream && (responseData.headers.contains("Grip-Hold") || acceptTypes.contains(contentType)))
+				if(acceptContentTypes.contains(contentType))
+				{
+					doAccept = true;
+				}
+				else
+				{
+					foreach(const HttpHeader &h, responseData.headers)
+					{
+						foreach(const QByteArray &hp, acceptHeaderPrefixes)
+						{
+							if(qstrnicmp(h.first.data(), hp.data(), hp.length()) == 0)
+							{
+								doAccept = true;
+								break;
+							}
+						}
+
+						if(doAccept)
+							break;
+					}
+				}
+			}
+
+			if(doAccept)
 			{
 				if(!buffering)
 				{
@@ -670,7 +740,7 @@ public slots:
 		if(sessionItems.isEmpty())
 		{
 			log_debug("proxysession: %p finished by passthrough", q);
-			emit q->finishedByPassthrough();
+			emit q->finished();
 		}
 	}
 
@@ -698,9 +768,11 @@ public slots:
 
 		if(allPaused)
 		{
-			AcceptData adata;
+			assert(!acceptRequest);
 
-			QList<RequestSession*> toDestroy;
+			responseData.body = responseBody.take();
+
+			AcceptData adata;
 
 			foreach(SessionItem *si, sessionItems)
 			{
@@ -712,41 +784,35 @@ public slots:
 				areq.peerAddress = si->rs->peerAddress();
 				areq.autoCrossOrigin = si->rs->autoCrossOrigin();
 				areq.jsonpCallback = si->rs->jsonpCallback();
+				areq.jsonpExtendedResponse = si->rs->jsonpExtendedResponse();
 				areq.inSeq = ss.inSeq;
 				areq.outSeq = ss.outSeq;
 				areq.outCredits = ss.outCredits;
 				areq.userData = ss.userData;
 				adata.requests += areq;
-
-				toDestroy += si->rs;
 			}
 
-			sessionItems.clear();
-			sessionItemsBySession.clear();
-
-			QPointer<QObject> self = this;
-			foreach(RequestSession *rs, toDestroy)
-			{
-				if(self) // <-- weird!
-					emit q->requestSessionDestroyed(rs, true);
-				delete rs;
-			}
-			if(!self)
-				return;
-
 			adata.requestData = requestData;
 			adata.requestData.body = requestBody.take();
 
 			adata.haveResponse = true;
 			adata.response = responseData;
-			adata.response.body = responseBody.take();
 
-			adata.route = routeId;
-			adata.channelPrefix = channelPrefix;
+			if(haveInspectData)
+			{
+				adata.haveInspectData = true;
+				adata.inspectData.doProxy = idata.doProxy;
+				adata.inspectData.sharingKey = idata.sharingKey;
+				adata.inspectData.userData = idata.userData;
+			}
 
-			log_debug("proxysession: %p finished for accept", q);
-			cleanup();
-			emit q->finishedForAccept(adata);
+			adata.route = route.id;
+			adata.channelPrefix = route.prefix;
+			adata.useSession = route.session;
+
+			acceptRequest = new AcceptRequest(acceptManager, this);
+			connect(acceptRequest, SIGNAL(finished()), SLOT(acceptRequest_finished()));
+			acceptRequest->start(adata);
 		}
 	}
 
@@ -767,12 +833,75 @@ public slots:
 
 		// don't destroy the RequestSession here. a finished signal will arrive next.
 	}
+
+	void acceptRequest_finished()
+	{
+		if(acceptRequest->success())
+		{
+			AcceptRequest::ResponseData rdata = acceptRequest->result();
+
+			delete acceptRequest;
+			acceptRequest = 0;
+
+			if(rdata.accepted)
+			{
+				// the requests were paused, so deleting them will leave the peer sessions active
+
+				QList<RequestSession*> toDestroy;
+				foreach(SessionItem *si, sessionItems)
+					toDestroy += si->rs;
+
+				sessionItems.clear();
+				sessionItemsBySession.clear();
+
+				QPointer<QObject> self = this;
+				foreach(RequestSession *rs, toDestroy)
+				{
+					emit q->requestSessionDestroyed(rs, true);
+					delete rs;
+					if(!self)
+						return;
+				}
+
+				log_debug("proxysession: %p finished for accept", q);
+				cleanup();
+				emit q->finished();
+			}
+			else
+			{
+				// wake up receivers
+				foreach(SessionItem *si, sessionItems)
+				{
+					si->state = SessionItem::WaitingForResponse;
+					si->rs->resume();
+				}
+
+				if(rdata.response.code != -1)
+					respondAll(rdata.response.code, rdata.response.reason, rdata.response.headers, rdata.response.body);
+				else
+					cannotAcceptAll();
+			}
+		}
+		else
+		{
+			delete acceptRequest;
+			acceptRequest = 0;
+
+			// wake up receivers and reject
+			foreach(SessionItem *si, sessionItems)
+			{
+				si->state = SessionItem::WaitingForResponse;
+				si->rs->resume();
+			}
+			cannotAcceptAll();
+		}
+	}
 };
 
-ProxySession::ProxySession(ZhttpManager *zhttpManager, DomainMap *domainMap, QObject *parent) :
+ProxySession::ProxySession(ZRoutes *zroutes, ZrpcManager *acceptManager, QObject *parent) :
 	QObject(parent)
 {
-	d = new Private(this, zhttpManager, domainMap);
+	d = new Private(this, zroutes, acceptManager);
 }
 
 ProxySession::~ProxySession()
@@ -780,9 +909,9 @@ ProxySession::~ProxySession()
 	delete d;
 }
 
-QByteArray ProxySession::routeId() const
+void ProxySession::setRoute(const DomainMap::Entry &route)
 {
-	return d->routeId;
+	d->route = route;
 }
 
 void ProxySession::setDefaultSigKey(const QByteArray &iss, const QByteArray &key)
